unit Kinf;

{$IFDEF FPC}
  {$MODE Delphi}
{$ENDIF}

interface

uses Classes, SysUtils, Crypto, DECCipher, DECHash, Tools, Eratosthenes,
     Generics.Collections;

type
  TKinfStorage = TDictionary<string, string>;

function GetKinfPath: string;
function ReadKinf(const FileName: string): TKinfStorage;

implementation

uses Windows, ShlObj, WinCrypt;

function GetKinfPath: string;
var
  AppData: array[0..MAX_PATH-1] of Char;
begin
  if SHGetFolderPath(0, CSIDL_LOCAL_APPDATA, 0, 0, @AppData) <> S_OK then
    Result := ''
  else
    Result := AppData + '\Amazon\Kindle\storage\.kinf2011';
end;

const
  KINF_HEADER_KEY_STR = 'n5Pr6St7Uv8Wx9YzAb0Cd1Ef2Gh3Jk4M';
  KINF_RECCOUNT_KEY_STR = 'AzB0bYyCeVvaZ3FfUuG4g-TtHh5SsIiR6rJjQq7KkPpL8lOoMm9Nn_c1XxDdW2wE';
  KINF_STOREITEM_KEY_STR = 'YvaZ3FfUm9Nn_c1XuG4yCAzB0beVg-TtHh5SsIiR6rJjQdW2wEq7KkPpL8lOoMxD';
  KINF_STORENAMES: array[0..16] of string = (
    'kindle.account.tokens','kindle.cookie.item','eulaVersionAccepted','login_date',
    'kindle.token.item','login','kindle.key.item','kindle.name.info','kindle.device.info',
    'MazamaRandomNumber', 'max_date', 'SIGVERIF',
    'build_version', 'SerialNumber', 'UsernameHash', 'kindle.directedid.info', 'DSN'
  );

function StrDecode(const Data, Key: string): RawByteString;
var
  i: Integer;
  High, Low: Byte;
begin
  Result := '';
  for i := 1 to Length(Data) div 2 do
  begin
    High := Pos(Data[i * 2 - 1], Key) - 1;
    Low := Pos(Data[i * 2], Key) - 1;
    if (High = 255) or (Low = 255) then
      Break;
    High := High * Length(Key);
    High := High xor $80;
    Result := Result + AnsiChar(High + Low);
  end;
end;

function StrEncode(const Data: RawByteString; const Key: string): string;
var
  i, L: Integer;
begin
  Result := '';
  L := Length(Key);
  for i := 1 to Length(Data) do
  begin
    Result := Result + Key[1 + (Ord(Data[i]) xor $80) div L];
    Result := Result + Key[1 + Ord(Data[i]) mod L];
  end;
end;

function EncodeHash(const Data, Key: string): string;
begin
  Result := StrEncode(THash_MD5.CalcBinary(RawByteString(Data)), Key);
end;

function UnprotectHeaderData(const Enc: RawByteString): string;
var
  Passwd, Salt, DecRaw: RawByteString;
  KeyIV: array[0..255] of Byte;
begin
  // Yes, let's use a key derivation function for a static password, cuz why not!
  Passwd := 'header_key_data';
  Salt := 'HEADER.2011';
  PBKDF2(Passwd[1], Length(Passwd), Salt[1], Length(Salt), @KeyIV[0], Length(KeyIV), $80);

  SetLength(DecRaw, Length(Enc));
  with TCipher_Rijndael.Create do
  begin
    Mode := cmCBCx;
    Init(KeyIV[0], 32, KeyIV[32], 16);
    Decode(Enc[1], DecRaw[1], Length(Enc));
    Free;
  end;
  Result := string(DecRaw);
end;

function ReadKinf(const FileName: string): TKinfStorage;
var
  SL: TStringList;
  Header, Build, Guid, KeyHash, EncData, Prefix, KeyName, Name: string;
  AddedEntropy, Enc: RawByteString;
  BlobProt, BlobEntropy, BlobUnprot: DATA_BLOB;
  i, j, RecCount, NOff: Integer;
begin
  Result := TKinfStorage.Create;
  SL := TStringList.Create;
  try
    SL.LoadFromFile(FileName);
    SL.Delimiter := '/';
    SL.StrictDelimiter := True;
    SL.DelimitedText := SL.Text;

    Header := UnprotectHeaderData(StrDecode(SL[0], KINF_HEADER_KEY_STR));

    if not Find(Header, '[Build:', ']', Build) or not Find(Header, '[Guid:', ']', Guid) then
      raise Exception.Create('Header data not found'#13#10 + Header);

    AddedEntropy := RawByteString(Build + Guid);

    i := 1;
    while i < SL.Count - 1 do
    begin
      KeyHash := Copy(SL[i], 1, 32);
      RecCount := StrToInt(string(StrDecode(Copy(SL[i], 35, 16), KINF_RECCOUNT_KEY_STR)));

      KeyName := 'unknown' + IntToStr(i);
      for Name in KINF_STORENAMES do
        if EncodeHash(Name, KINF_STOREITEM_KEY_STR) = KeyHash then
        begin
          KeyName := Name;
          Break;
        end;

      EncData := '';
      for j := 1 to RecCount do
      begin
        Inc(i);
        EncData := EncData + SL[i];
      end;

      with TSieveOfEratosthenes.Create(Length(EncData) div 3) do
      begin
        NOff := Cardinal(Length(EncData)) - Primes[PrimesFound - 1];
        Free;
      end;
      Prefix := Copy(EncData, 1, NOff);
      Delete(EncData, 1, NOff);
      EncData := EncData + Prefix;

      Enc := StrDecode(EncData, KINF_STOREITEM_KEY_STR);

      SetDataBlobText(Enc, BlobProt);
      SetDataBlobText(THash_SHA1.CalcBinary(RawByteString(KeyHash)) + AddedEntropy, BlobEntropy);
      try
        if not CryptUnprotectData(@BlobProt, nil, @BlobEntropy, nil, nil, 1, @BlobUnprot) then
          RaiseLastOSError;

        Result.Add(KeyName, GetDataBlobText(BlobUnprot));
      finally
        FreeDataBlob(BlobProt);
        FreeDataBlob(BlobEntropy);
        FreeDataBlob(BlobUnprot);
      end;

      Inc(i);
    end;
  finally
    SL.Free;
  end;
end;

end.
